#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# License: GNU General Public License v2
#
# Author: thl-cmk[at]outlook[dot]com
# URL   : https://thl-cmk.hopto.org
# Date  : 2024-07-05
# File  : vsphere_topo/agent_based/packages.py

# 2024-07-10: added time taken as metric
# 2024-07-11: added clearer error message for missing section_esx_vsphere_clusters section
#             added automatic workaround for backend is only picking up topologies from default folder
#             -> obsoletes 'Add dummy topologies'
#             added vCenter twice (once as vCenter once as vM) with different object IDs
#             -> obsoletes 'Don\'t add vCenter as VM'
# 2024-07-12: added support for Queried host is a host system
# 2024-07-13: refactoring save topology/make default and add dummy's logic
# 2024-07-16: moved add_host/add_service/add_connection to lib/utils/TopoObjects/TopoConnections
#             moved ruleset name constants to utils/ruleset_names.py
# 2024-07-25: fixed crash if VM has no data store
# 2024-07-29: added option to ignore powered off VMs
#             added option to ignore VMs by name (regex)
# 2024-07-31: added option to map vSphere names to CMK/topology names for data center, cluster, hosts and VMs

from collections.abc import Mapping, Sequence
from re import compile as re_compile
from time import time_ns


from cmk.agent_based.v2 import (
    CheckPlugin,
    CheckResult,
    DiscoveryResult,
    Result,
    Service,
    State,
    check_levels,
    render,
)
from cmk.base.check_api import host_name
from cmk_addons.plugins.vsphere_topo.constants import (
    EMBLEM_CLUSTER,
    EMBLEM_DATA_CENTER,
    EMBLEM_DATA_STORE,
    ICON_VCENTER,
    PARAM_CLUSTER,
    PARAM_DATA_CENTER,
    PARAM_DATA_STORE,
    PARAM_DATA_STORE_AS_SERVICE,
    PARAM_HOST_SYSTEMS,
    PARAM_IGNORE_POWERED_OFF_VMS,
    PARAM_IGNORE_VM_NAME_REGEX,
    PARAM_MAKE_DEFAULT,
    PARAM_MAP_NAMES,
    PARAM_VCENTER,
    PARAM_VM_NAMES,
    RULE_SET_NAME_VSPHERE_TOPO,
    TOPOLOGY_NAME,
)
from cmk_addons.plugins.vsphere_topo.lib.utils import (
    BASE_TOPO_PATH,
    LiveStatusConnection,
    TopoConnections,
    TopoObjects,
    add_dummy_topologies,
    adjust_name,
    get_emblem,
    get_name_map,
    make_topo_default,
    save_topology,
)

Section = Mapping[str, object]


def discover_vsphere_topo(
        section_esx_vsphere_clusters: Section | None,
        section_esx_vsphere_virtual_machines: Sequence[Section] | None,
        section_esx_vsphere_datastores: Section | None,
) -> DiscoveryResult:
    yield Service()


def check_vsphere_topo(
        params: Mapping[str, object],
        section_esx_vsphere_clusters: Section | None,
        section_esx_vsphere_virtual_machines: Sequence[Section] | None,
        section_esx_vsphere_datastores: Section | None,
) -> CheckResult:
    start_time = time_ns()
    objects: TopoObjects = TopoObjects()
    connections: TopoConnections = TopoConnections()
    ignored_powered_off_vms: int = 0
    ignored_by_name_vms: int = 0

    name_map = get_name_map(params.get(PARAM_MAP_NAMES, []))

    if ignore_by_name_str := ')|('.join(params.get(PARAM_IGNORE_VM_NAME_REGEX, [])):
        re_ignore_vms = re_compile(f'({ignore_by_name_str})')
    else:
        re_ignore_vms = None

    if section_esx_vsphere_clusters is not None:
        __clusters = {
            'HA': {
                'datacenter': 'datacenter-21',
                'hostsystems': 'esxi01, esxi02',
                'vms': 'vmname01, vmname02, vmname03'
            }
        }

        raw_vsphere_host: str = host_name().strip()
        objects.add_host(
            host=raw_vsphere_host,
            obj_id_prefix='vc',
            icon=get_emblem(ICON_VCENTER, params.get(PARAM_VCENTER)),
        )
        vsphere_host = f'vc{raw_vsphere_host}'

        for raw_cluster in section_esx_vsphere_clusters:
            data_center: str = section_esx_vsphere_clusters[raw_cluster]['datacenter'].strip()
            data_center = name_map.get(data_center, data_center)
            if objects.topo_objects.get(data_center) is None:
                objects.add_host(
                    host=data_center,
                    emblem=get_emblem(EMBLEM_DATA_CENTER, params.get(PARAM_DATA_CENTER)),
                    link2core=False,
                )
                connections.add_connection(data_center, vsphere_host)
            cluster = name_map.get(raw_cluster, raw_cluster)
            objects.add_host(
                host=cluster,
                emblem=get_emblem(EMBLEM_CLUSTER, params.get(PARAM_CLUSTER)),
                link2core=False
            )
            connections.add_connection(cluster, data_center)

            host_systems: Sequence[str] = section_esx_vsphere_clusters[raw_cluster].get('hostsystems', '').split(',')
            for host_system in host_systems:
                esx_name: str = adjust_name(host_system, params.get(PARAM_HOST_SYSTEMS, {}))
                esx_name = name_map.get(esx_name, esx_name)
                objects.add_host(host=esx_name)
                connections.add_connection(cluster, esx_name)
        yield Result(
            state=State.OK,
            notice='VMware ESX via vSphere agent type of query: Queried host is the vCenter'
        )
    else:  # assuming the query type is ESXi host
        raw_vsphere_host: str = host_name().strip()
        objects.add_host(host=raw_vsphere_host)
        vsphere_host = raw_vsphere_host
        yield Result(
            state=State.OK,
            notice='VMware ESX via vSphere agent type of query: Queried host is a host system'
        )

    if section_esx_vsphere_virtual_machines is not None:
        __VMs = [
            {
                'vm_name': 'vmname01',
                'hostsystem': 'esxi01',
                'powerstate': 'poweredOff',
                'guest_os': 'Microsoft Windows Server 2008 R2 (64-bit)',
                'compatibility': 'vmx-10',
                'uuid': '420354bd-bee0-88d5-053c-d4f424106b0f'
            },
            {
                'vm_name': 'vmname02',
                'hostsystem': 'esxi02',
                'powerstate': 'poweredOn',
                'guest_os': 'Microsoft Windows Server 2019 (64-bit)',
                'compatibility': 'vmx-19',
                'uuid': '422eba9f-0327-2da0-b546-f5a5e5dfdb68'
            },
            {
                'vm_name': 'vmname03',
                'hostsystem': 'esxi01',
                'powerstate': 'poweredOn',
                'guest_os': 'Other 3.x or later Linux (64-bit)',
                'compatibility': 'vmx-11',
                'uuid': '422e3863-b7de-dc55-74f8-6dc9c159b36d'
            }
        ]
        for vm in section_esx_vsphere_virtual_machines:
            if params.get(PARAM_IGNORE_POWERED_OFF_VMS) is not True or vm.get('powerstate') != 'poweredOff':
                vm_name: str = adjust_name(vm['vm_name'], params.get(PARAM_VM_NAMES, {}))
                if re_ignore_vms is None or re_ignore_vms.match(vm_name) is None:
                    vm_name = name_map.get(vm_name, vm_name)
                    host_system: str = adjust_name(vm['hostsystem'], params.get(PARAM_HOST_SYSTEMS, {}))
                    host_system = name_map.get(host_system, host_system)
                    objects.add_host(host=vm_name)
                    connections.add_connection(vm_name, host_system)
                else:
                    ignored_by_name_vms += 1
            else:
                ignored_powered_off_vms += 1

    service_prefix: str = ''
    ds_services = None
    ls_connection = LiveStatusConnection()

    if section_esx_vsphere_datastores is not None:
        data_stores_host: str = 'data_stores'  # Anchor for data stores
        objects.add_host(
            host=data_stores_host,
            link2core=False,
            emblem=get_emblem(EMBLEM_DATA_STORE, params.get(PARAM_DATA_STORE)),
        )
        for ds_name in section_esx_vsphere_datastores:
            objects.add_host(
                host=ds_name,
                link2core=False,
                emblem=get_emblem(EMBLEM_DATA_STORE, params.get(PARAM_DATA_STORE)),
            )
            connections.add_connection(ds_name, data_stores_host)

        if params.get(PARAM_DATA_STORE_AS_SERVICE) is True:
            query: str = (
                'GET services\n'
                'Columns: service_description\n'
                'Filter: service_check_command ~ ^check_mk-esx_vsphere_datastores(!.*)?\n'
                f'Filter: host_name = {vsphere_host}\n'
                'OutputFormat: python3\n'
            )
            if (ds_services := ls_connection.query(query=query)) is not None:
                for ds_service in ds_services:
                    service_name: str = ds_service[0]
                    service_id: str = f'{service_name}@{vsphere_host}'
                    ds_name: str = service_name.split('fs_')[-1].split('Filesystem ')[-1]
                    objects.add_service(
                        host=vsphere_host,
                        service=service_name,
                    )
                    connections.add_connection(service_id, ds_name)
                    service_prefix: str = service_name.split(ds_name)[0]

    query: str = (
        'GET services\n'
        'Columns: host_name long_plugin_output\n'
        'Filter: description = ESX Datastores\n'
        'OutputFormat: python3\n'
    )
    if (vm_to_data_stores := ls_connection.query(query=query)) is not None:
        for vm, data_stores in vm_to_data_stores:
            vm_name: str = adjust_name(vm, params.get(PARAM_VM_NAMES, {}))
            if vm_name in objects.topo_objects:
                data_stores = data_stores.split(',')
                for data_store in data_stores:
                    try:
                        ds_name = data_store.strip().split(' ')[2]
                    except IndexError:
                        continue
                    else:
                        if ds_services is not None:
                            service_name = f'{service_prefix}{ds_name}'
                            service_id = f'{service_name}@{vsphere_host}'
                            connections.add_connection(service_id, vm_name)
                        else:
                            connections.add_connection(ds_name, vm_name)

    connections.topo_connections.sort()
    data_set = {
        'version': 1,
        'name': TOPOLOGY_NAME,
        'objects': dict(sorted(objects.topo_objects.items())),
        'connections': connections.topo_connections,
    }

    save_topology(data=data_set, sub_directory=raw_vsphere_host)
    # workaround for backend is only picking up topologies from default folder
    add_dummy_topologies(sub_directory=raw_vsphere_host)
    # end workaround
    make_topo_default(sub_directory=raw_vsphere_host, make_default=params.get(PARAM_MAKE_DEFAULT, False))

    yield Result(state=State.OK, summary=f'Objects: {len(objects.topo_objects)}')
    yield Result(state=State.OK, summary=f'Connections: {len(connections.topo_connections)}')
    if params.get(PARAM_IGNORE_POWERED_OFF_VMS) is True:
        yield Result(state=State.OK, summary=f'Ignored powered off VMs: {ignored_powered_off_vms}')
    if re_ignore_vms is not None:
        yield Result(state=State.OK, summary=f'Ignored VMs by name: {ignored_by_name_vms}')
    yield Result(
        state=State.OK,
        notice=f'Written to: {BASE_TOPO_PATH}/{raw_vsphere_host}/data_{TOPOLOGY_NAME.lower()}.json',
    )
    yield from check_levels(
        value=(time_ns() - start_time) / 1e9,
        label='Time taken',
        metric_name=f'{TOPOLOGY_NAME.lower()}_topo_time_taken',
        boundaries=(0, None),
        render_func=render.timespan,
    )


check_plugin_vsphere_topo = CheckPlugin(
    name='vsphere_topo',
    service_name=f'{TOPOLOGY_NAME} Topology',
    sections=[
        'esx_vsphere_clusters',
        'esx_vsphere_virtual_machines',
        'esx_vsphere_datastores',
    ],
    check_function=check_vsphere_topo,
    discovery_function=discover_vsphere_topo,
    check_ruleset_name=RULE_SET_NAME_VSPHERE_TOPO,
    check_default_parameters={},
)
